﻿#define USE_APP_RUN

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.Runtime.InteropServices;
using System.Threading;
using System.Windows.Forms;

namespace CSClsLibEx
{
    public delegate void EventStep(object sender, StepEventArgs e);

    [ComVisible(true)]
    public class StepEventArgs : EventArgs
    {
        public int count { get; protected set; }
        public bool completed { get; protected set; }
        public StepEventArgs(int count, bool completed)
        {
            this.count = count;
            this.completed = completed;
        }
    }

    [ComVisible(true)]
    [Guid("C76AE6CD-2F96-4DB7-BB0C-A37392AC8176")]
    [InterfaceType(ComInterfaceType.InterfaceIsIDispatch)]
    public interface StepEvent
    {
        void EventStep(object sender, StepEventArgs args);
    }

    [ComVisible(true)]
    [Guid("89CE323D-979E-4154-A733-07C3105B7CDA")]
    [ProgId("CSClsLibEx.ClsSampleEx")]
    [ClassInterface(ClassInterfaceType.AutoDual)]
    [ComSourceInterfaces(typeof(StepEvent))]
    public class ClsSampleEx
    {
        const int THREAD_SLEEP = 100;
        public string status { get; protected set; }
        public int maxCount { get; protected set; }
        public bool terminated { get; protected set; }
        public bool useAppRun { get; protected set; }

        public event EventStep eventStep;
        ProgressDlg progressDlg;
        ThreadInfo threadInfo;
        EventHandler threadCompleted;

        public void Terminate()
        {
            if (!this.terminated)
            {
                this.terminated = true;
            }
        }

        public void RunWork([MarshalAs(UnmanagedType.SafeArray, SafeArraySubType = VarEnum.VT_VARIANT)] 
            object[] arg)
        {
            this.terminated = false;
            this.maxCount = Convert.ToInt32(arg[1]);
            this.eventStep += new EventStep(ClsSampleEx_eventStep);
            this.threadInfo = new ThreadInfo(arg);
            this.threadInfo.thread = new Thread(RunWorkEx);
            this.threadCompleted = new EventHandler(ThreadCompleted);
            this.threadInfo.thread.SetApartmentState(ApartmentState.STA);  
#if USE_APP_RUN
            this.progressDlg = new ProgressDlg(this);
            this.progressDlg.Show();
            this.useAppRun = true;
            Application.DoEvents(); 
#endif
            this.threadInfo.thread.Start(threadInfo);
            while (!this.threadInfo.thread.IsAlive)
            {
                Thread.Sleep(0);
            }
            if (this.useAppRun && this.progressDlg != null)
            {
                Application.Run(this.progressDlg);
            }
        }

        void ClsSampleEx_eventStep(object sender, StepEventArgs e)
        {
            if (this.progressDlg != null)
            {
                this.progressDlg.ResetProgressBar(
                    sender, 
                    (100 * (e.count + 1)) / this.maxCount, 
                    e.completed);
            }
        }

        protected void OnEventStep(StepEventArgs args)
        {
            if (this.eventStep != null)
            {
                this.eventStep(this, args);
            }
        }

        void RunWorkEx(object objThreadInfo)
        {
            int count = 0;
            int maxCount = (-1);
            ThreadInfo threadInfo = null;
 
            try
            {
                if ((threadInfo = objThreadInfo as ThreadInfo) == null)
                {
                    throw new NullReferenceException("threadInfo"); 
                }
                if (String.IsNullOrEmpty(this.status = Convert.ToString(threadInfo.arg[0])))
                {
                    throw new ArgumentNullException("threadInfo.arg[0]");
                }
                if ((maxCount = Convert.ToInt32(threadInfo.arg[1])) <= 0)
                {
                    throw new ArgumentOutOfRangeException("threadInfo.arg[1] = " + maxCount);
                }
                if (this.progressDlg == null)
                {
                    this.progressDlg = new ProgressDlg(this);
                    this.progressDlg.Show();
                    Application.DoEvents();
                }
                if (this.progressDlg == null)
                {
                    throw new NullReferenceException("this.progressDlg");
                }
                for (; count < maxCount; count++)
                {
                    this.OnEventStep(new StepEventArgs(count, false));
                    Thread.Sleep(THREAD_SLEEP);
                    if (this.terminated
                        || this.progressDlg == null)
                    {
                        break;
                    }
                }
            }
            finally
            {
                if (this.progressDlg != null)
                {
                    if (this.progressDlg.InvokeRequired)
                    {
                        this.progressDlg.CloseEx(this);
                    }
                    else
                    {
                        this.progressDlg.Close();
                    }
                }
                this.OnEventStep(new StepEventArgs(count, true));
                this.threadCompleted.BeginInvoke(this, new EventArgs(), null, null);
            }
        }

        void ThreadCompleted(object sender, EventArgs e)
        {
            while (this.threadInfo.thread != null
                && this.threadInfo.thread.IsAlive)
            {
                Thread.Sleep(0);
            }
            if (this.threadInfo.thread != null)
            {
                this.threadInfo.thread.Join();
                this.threadInfo.thread = null; 
            }
            this.eventStep -= new EventStep(ClsSampleEx_eventStep);
        }
    }

    public class ThreadInfo
    {
        public object[] arg { get; protected set; }
        public Thread thread { get; set; }

        public ThreadInfo(object[] arg)
        {
            this.arg = new object[arg.Length]; 
            arg.CopyTo(this.arg, 0);
        }
    }
}
